#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <sys/utsname.h>

#include "dosemu.h"
#include "arch/signal.h"

/*
 * DANG_BEGIN_FUNCTION version_init
 * 
 * description:
 *  Find version of OS running and set necessary global parms.
 *
 * DANG_END_FUNCTION
 */
static void version_init(void)
{
  struct utsname unames;

  uname(&unames);
  warn("DOSEMU-%s is coming up on %s version %s\n", VERSTR, unames.sysname,
       unames.release);
  warn("Built for %d\n", KERNEL_VERSION);
  if (unames.release[0] > 0 ) {
    if ((unames.release[2] == 1  && unames.release[3] > 1 ) ||
         unames.release[2] > 1 ) {
      use_sigio=FASYNC;
    }
  }

  /* Next Check input */
  if (isatty(STDIN_FILENO)) {
    k_printf("STDIN is tty\n");
    config.kbd_tty = 0;
  } else {
    k_printf("STDIN not a tty\n");
    config.kbd_tty = 1;
  }
}

/*
 * DANG_BEGIN_FUNCTION io_select_init
 * 
 * description:
 * Initialize fd_sets to NULL for both SIGIO and NON-SIGIO.
 * 
 * DANG_END_FUNCTION
 */
static void io_select_init(void)
{
  FD_ZERO(&fds_sigio);	/* initialize both fd_sets to 0 */
  FD_ZERO(&fds_no_sigio);
}

void arch_init(void)
{
  version_init();
  io_select_init();
  linux_signal_init();
}

void arch_term(void)
{
}
