/*
 * DANG_BEGIN_MODULE
 *
 * Description:
 *   This module contains the BIOS default interrupt handlers for DOSEMU.
 *   Most of these functions will never be invoked (at least in the i386
 *   versions) because they will be emulated in the DOSEMU BIOS itself.
 *
 * DANG_END_MODULE
 *
 * DANG_BEGIN_CHANGELOG
 *
 * $Date: $
 * $Source: $
 * $Revision: $
 * $State: $
 *
 * $Log: $
 *
 * DANG_END_CHANGELOG
 */

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <sys/wait.h>
#include <sys/time.h>

#include "dosemu.h"
#include "emu_int.h"
#include "emu_bios.h"
#include "emu_port.h"
#include "emu_mem.h"
#include "emu_timers.h"
#include "emu_mouse.h"

static void int08(Bit8u i)
{
  Bit32u ticks;
  g_printf("BIOS: int08 timer tick\n");

  ticks = READ_DWORD(bios_timer);
  if (++ticks >= 0x001b001b) {
    ticks -= 0x001b001b;
    WRITE_BYTE(bios_timer_overflow, READ_BYTE(bios_timer_overflow) + 1);
  }
  WRITE_DWORD(bios_timer, ticks);
  __cli__;
  int_run(0x1c);
  port_outb(0x20, 0x20);
}

static void int09(Bit8u i)
{
  g_printf("BIOS: int09 keyboard\n");
  /* what to do here??? */
  port_outb(0x20, 0x20);
}

static void int_irq_lo(Bit8u i)
{
  g_printf("BIOS: irq %d reached dosemu...\n", i-0x08);
  port_outb(0x20, 0x20);
}

static void int11(Bit8u i)
{
  _AX = READ_WORD(BIOS_CONFIGURATION);
}

static void int12(Bit8u i)
{
  _AX = READ_WORD(BIOS_MEMORY_SIZE);
}

static void int15(Bit8u i)
{
  static struct timeval scr_tv;
  struct timeval wait_time;
  int num;

  if (_AH != 0x4f)
    __clc__;
  WRITE_FLAGS(READ_FLAGS() | VIF);

  switch (_AH) {
  case 0x10:			/* TopView/DESQview */
    switch (_AL)
    {
    case 0x00:			/* giveup timeslice */
      usleep(INT15_IDLE_USECS);
      return;
    }
    __stc__;
    break;
  case 0x41:			/* wait on external event */
    break;
  case 0x4f:			/* Keyboard intercept */
    _AH = 0x86;
    k_printf("INT15 0x4f CARRY=%x AX=%lx\n", (WORD(READ_FLAGS()) & CF), _AX);
/*
    __stc__;
    if (_AL & 0x80 )
      if (1 || !(_AL&0x80) ){
	fprintf(stderr, "Carrying it out\n");
        __stc__;
      }
      else
	__clc__;
*/
    break;
  case 0x80:			/* default BIOS device open event */
    LWORD(eax) &= 0x00FF;
    return;
  case 0x81:
    LWORD(eax) &= 0x00FF;
    return;
  case 0x82:
    LWORD(eax) &= 0x00FF;
    return;
  case 0x83:
    h_printf("int 15h event wait:\n");
    show_regs(__FILE__, __LINE__);
    __stc__;
    return;			/* no event wait */
  case 0x84:
    __stc__;
    return;			/* no joystick */
  case 0x85:
    num = LWORD(eax) & 0xFF;	/* default bios handler for sysreq key */
    if (num == 0 || num == 1) {
      LWORD(eax) &= 0x00FF;
      return;
    }
    LWORD(eax) &= 0xFF00;
    LWORD(eax) |= 1;
    __stc__;
    return;
  case 0x86:
    /* wait...cx:dx=time in usecs */
    g_printf("doing int15 wait...ah=0x86\n");
    show_regs(__FILE__, __LINE__);
    wait_time.tv_sec = 0;
    wait_time.tv_usec = (LWORD(ecx) << 16) | LWORD(edx);
    RPT_SYSCALL(select(STDIN_FILENO, NULL, NULL, NULL, &scr_tv));
    __clc__;
    return;

  case 0x87:
    if (config.xms_size)
      xms_int15();
    else {
      LWORD(eax) &= 0xFF;
      LWORD(eax) |= 0x0300;	/* say A20 gate failed - a lie but enough */
      __stc__;
    }
    return;

  case 0x88:
    if (config.xms_size) {
      xms_int15();
    }
    else {
      LWORD(eax) &= ~0xffff;	/* no extended ram if it's not XMS */
      __clc__;
    }
    return;

  case 0x89:			/* enter protected mode : kind of tricky! */
    LWORD(eax) |= 0xFF00;		/* failed */
    __stc__;
    return;
  case 0x90:			/* no device post/wait stuff */
    __stc__;
    return;
  case 0x91:
    __stc__;
    return;
  case 0xc0:
    LWORD(es) = ROM_CONFIG_SEG;
    LWORD(ebx) = ROM_CONFIG_OFF;
    _AL = 0;
    return;
  case 0xc1:
    __stc__;
    return;			/* no ebios area */
  case 0xc2:
    m_printf("PS2MOUSE: Call ax=0x%04x\n", LWORD(eax));
    if (!mice->intdrv)
      if (mice->type != MOUSE_PS2) {
	REG(eax) = 0500;        /* No ps2 mouse device handler */
	__stc__;
	return;
      }
                
    switch (REG(eax) &= 0x00FF)
      {
      case 0x0000:                    
	mouse.ps2.state = _BH;
	if (mouse.ps2.state == 0)
	  mice->intdrv = FALSE;
	else
	  mice->intdrv = TRUE;
 	_AL = 0;
	__clc__;		
 	break;
      case 0x0001:
	_AL = 0;
	_BX = 0xAAAA;    /* we have a ps2 mouse */
	__clc__;
	break;
      case 0x0003:
	if (_BH != 0) {
	  __stc__;
	  _AH = 1;
	} else {
	  __clc__;
	  _AH = 0;
	}
	break;
      case 0x0004:
	HI(bx) = 0xAA;
	_AH = 0;
	__clc__;
	break;
      case 0x0005:			/* Initialize ps2 mouse */
	_AH = 0;
	mouse.ps2.pkg = HI(bx);
	__clc__;
	break;
      case 0x0006:
	switch (HI(bx)) {
	case 0x00:
	  LO(bx)  = (mouse.rbutton ? 1 : 0);
	  LO(bx) |= (mouse.lbutton ? 4 : 0);
	  LO(bx) |= 0; 	/* scaling 1:1 */
	  LO(bx) |= 0x20;	/* device enabled */
	  LO(bx) |= 0;	/* stream mode */
	  LO(cx) = 0;		/* resolution, one count */
	  LO(dx) = 0;		/* sample rate */
	  _AH = 0;
	  __clc__;
	  break;
	case 0x01:
	  _AH = 0;
	  __clc__;
	  break;
	case 0x02:
	  _AH = 1;
	  __stc__;
	  break;
	}
	break;
#if 0
      case 0x0007:
	pushw(ssp, sp, 0x000B);
	pushw(ssp, sp, 0x0001);
	pushw(ssp, sp, 0x0001);
	pushw(ssp, sp, 0x0000);
	REG(cs) = REG(es);
	REG(eip) = REG(ebx);
	_AH = 0;
	__clc__;
	break;
#endif
     default:
	_AH = 1;
	g_printf("PS2MOUSE: Unknown call ax=0x%04x\n", LWORD(eax));
	__stc__;
      }
    return;
  case 0xc3:
    /* no watchdog */
    __stc__;
    return;
  case 0xc4:
    /* no post */
    __stc__;
    return;
  default:
    g_printf("int 15h error: ax=0x%04x\n", LWORD(eax));
    __stc__;
    return;
  }
}

static void int16(Bit8u i)
{
  /* what to do here??? */
}

/* int17 -- printer, handled elsewhere */

static void int18(Bit8u i)
{
  g_printf("BASIC interrupt being attempted.\n");
}

static void int19(Bit8u i)
{
  boot();
}

static void int1a(Bit8u i)
{
  unsigned int test_date;
  time_t akt_time; 
  time_t time_val;
  struct tm *tm;

  switch (_AH) {

    /* A timer read should reset the overflow flag */
  case 0:			/* read time counter */
    time(&akt_time);
    tm = localtime((time_t *) &akt_time);
    test_date = tm->tm_year * 10000 + tm->tm_mon * 100 + tm->tm_mday;
    if ( check_date != test_date ) {
      start_time = akt_time;
      check_date = tm->tm_year * 10000 + tm->tm_mon * 100 + tm->tm_mday;
      g_printf("Over 24hrs forward or backward\n");
      WRITE_BYTE(BIOS_24_HOURS_FLAG, READ_BYTE(BIOS_24_HOURS_FLAG)+1);
    }
    last_ticks = (tm->tm_hour * 60 * 60 + tm->tm_min * 60 + tm->tm_sec)*18.206;
    set_ticks(last_ticks);

    _AL = READ_BYTE(BIOS_24_HOURS_FLAG);
    _CX = (READ_DWORD(BIOS_TIMER) >> 16) & 0xffff;
    _DX = READ_DWORD(BIOS_TIMER) & 0xffff;
#if 0
    g_printf("read timer st:%u ticks:%u act:%u, actdate:%d\n",
	     start_time, last_ticks, akt_time, tm->tm_mday);
#endif
    break;
  case 1:			/* write time counter */
    last_ticks = (LWORD(ecx) << 16) | (LWORD(edx) & 0xffff);
    set_ticks(last_ticks);
    time(&start_time);
    g_printf("set timer to %lu \n", last_ticks);
    break;
  case 2:			/* get time */
    time(&time_val);
    tm = localtime((time_t *) &time_val);
    g_printf("get time %d:%02d:%02d\n", tm->tm_hour, tm->tm_min, tm->tm_sec);
    HI(cx) = tm->tm_hour % 10;
    tm->tm_hour /= 10;
    HI(cx) |= tm->tm_hour << 4;
    LO(cx) = tm->tm_min % 10;
    tm->tm_min /= 10;
    LO(cx) |= tm->tm_min << 4;
    HI(dx) = tm->tm_sec % 10;
    tm->tm_sec /= 10;
    HI(dx) |= tm->tm_sec << 4;
    /* LO(dx) = tm->tm_isdst; */
    /* REG(eflags) &= ~CF; */
    WRITE_FLAGS(READ_FLAGS() & ~CF);
    break;
  case 4:			/* get date */
    time(&time_val);
    tm = localtime((time_t *) &time_val);
    tm->tm_year += 1900;
    tm->tm_mon++;
    g_printf("get date %02d.%02d.%04d\n", tm->tm_mday, tm->tm_mon, tm->tm_year);
#if 0
    gettimeofday(&tp, &tzp);
    ticks = tp.tv_sec - (tzp.tz_minuteswest * 60);
#endif
    LWORD(ecx) = tm->tm_year % 10;
    tm->tm_year /= 10;
    LWORD(ecx) |= (tm->tm_year % 10) << 4;
    tm->tm_year /= 10;
    LWORD(ecx) |= (tm->tm_year % 10) << 8;
    tm->tm_year /= 10;
    LWORD(ecx) |= (tm->tm_year) << 12;
    LO(dx) = tm->tm_mday % 10;
    tm->tm_mday /= 10;
    LO(dx) |= tm->tm_mday << 4;
    HI(dx) = tm->tm_mon % 10;
    tm->tm_mon /= 10;
    HI(dx) |= tm->tm_mon << 4;
    /* REG(eflags) &= ~CF; */
    WRITE_FLAGS(READ_FLAGS() & ~CF);
    break;
  case 3:			/* set time */
  case 5:			/* set date */
    g_printf("ERROR: timer: can't set time/date\n");
    break;
  default:
    g_printf("ERROR: timer error AX=0x%04x\n", LWORD(eax));
    break;
  }
}

static void int70(Bit8u i)
{
  g_printf("BIOS: int75 -- irq8 (rtc) unhandled by DOSEMU\n");
  port_outb(0x80, 0x20);
  port_outb(0x20, 0x20);
}

static void int_irq_hi(Bit8u i)
{
  g_printf("BIOS: irq %d unhandled by DOSEMU\n", i-0x70);
  port_outb(0x80, 0x20);
  port_outb(0x20, 0x20);
}

static void int75(Bit8u i)
{
  g_printf("BIOS: int75 -- irq13 unhandled by DOSEMU\n");
  port_outb(0x80, 0x20);
  port_outb(0x20, 0x20);
  /* Norton's SYSINFO says, "Redirect to NMI interrupt" ??? */
}

static void int77(Bit8u i)
{
  g_printf("BIOS: int77 -- irq15 -> redirect to irq2 unhandled\n");
  port_outb(0x80, 0x20);
  int_run(0x0a);
  /* is this correct? */
}

void bios_int_init(void)
{
}

void bios_int_reset(void)
{
  int i;
  /*
   * Note:  we could copy interrupt vectors 0x00-0x13 from the table
   *        located at 0xf000:0xfef3 if we really wanted to
   * Note:  The following will point to IRET unless someone else has
   *        a better idea...
   *           0,1,3,4 - processor errors
   *           18      - basic interpreter
   */
  g_printf("BIOS: setting vm86 int vectors\n");

  SETIVEC(0x00, BIOSSEG, ROM_BIOS_IRET);		/* divide by zero */
  SETIVEC(0x01, BIOSSEG, ROM_BIOS_IRET);		/* single step */
  SETIVEC(0x02, BIOSSEG, ROM_BIOS_NMI);		/* NMI */
  SETIVEC(0x03, BIOSSEG, ROM_BIOS_IRET);		/* breakpoint */
  SETIVEC(0x04, BIOSSEG, ROM_BIOS_IRET);		/* overflow */
  SETIVEC(0x05, BIOSSEG, ROM_BIOS_INT05);		/* print screen */
  SETIVEC(0x06, BIOSSEG, ROM_BIOS_IRET);		/* reserved */
  SETIVEC(0x07, BIOSSEG, ROM_BIOS_IRET);		/* reserved */
  SETIVEC(0x08, BIOSSEG, ROM_BIOS_INT08);		/* irq 0 - timer */
  SETIVEC(0x09, BIOSSEG, ROM_BIOS_INT09);		/* irq 1 - keyboard */
  for (i = 0x0A; i < 0x10; i++)
    SETIVEC(i, BIOSSEG, ROM_BIOS_IRQ_LO);		/* irq 2-7 */
  SETIVEC(0x10, BIOSSEG, ROM_BIOS_INT10);		/* video interrupt */
  SETIVEC(0x11, BIOSSEG, ROM_BIOS_INT11);		/* equipment */
  SETIVEC(0x12, BIOSSEG, ROM_BIOS_INT12);		/* memory */
  SETIVEC(0x13, BIOSSEG, ROM_BIOS_INT13);		/* disk */
  SETIVEC(0x14, BIOSSEG, ROM_BIOS_INT14);		/* serial */
  SETIVEC(0x15, BIOSSEG, ROM_BIOS_INT15);		/* system services */
  SETIVEC(0x16, BIOSSEG, ROM_BIOS_INT16);		/* keyboard */
  SETIVEC(0x17, BIOSSEG, ROM_BIOS_INT17);		/* printer */
  SETIVEC(0x18, BIOSSEG, ROM_BIOS_IRET);		/* resident basic */
  SETIVEC(0x19, BIOSSEG, ROM_BIOS_INT19);		/* bootstrap loader */
  SETIVEC(0x1A, BIOSSEG, ROM_BIOS_INT1A);		/* rtc services */
  SETIVEC(0x1B, BIOSSEG, ROM_BIOS_IRET);		/* keyboard break */
  SETIVEC(0x1C, BIOSSEG, ROM_BIOS_IRET);		/* user timer tick */
  SETIVEC(0x1D, BIOSSEG, ROM_BIOS_VIDEOPARM);	/* video parameters */
  SETIVEC(0x1E, BIOSSEG, ROM_BIOS_FLOPPARM);	/* diskette parameters */
  SETIVEC(0x1F, BIOSSEG, ROM_BIOS_8x8FONT);		/* video characters */
  for (i = 0x20; i < 0x3F; i++)
    SETIVEC(i, BIOSSEG, ROM_BIOS_IRET);		/* DOS interrupts */
  SETIVEC(0x40, BIOSSEG, ROM_BIOS_INT13FLOP);	/* relocated diskette */
  SETIVEC(0x41, BIOSSEG, ROM_BIOS_HARDPARM2);	/* fixed disk parameters */
  SETIVEC(0x42, BIOSSEG, ROM_BIOS_INT10);		/* relocated video */
  for (i = 0x43; i < 0x46; i++)
    SETIVEC(i, BIOSSEG, ROM_BIOS_IRET);		/* reserved */
  SETIVEC(0x46, BIOSSEG, ROM_BIOS_HARDPARM);	/* fixed disk parameters */
  for (i = 0x47; i < 0x60; i++)
    SETIVEC(i, BIOSSEG, ROM_BIOS_IRET);		/* reserved */
  for (i = 0x60; i < 0x66; i++)
    SETIVEC(i, 0x0000, 0x0000);			/* reserved for user progs */
  for (i = 0x67; i < 0x70; i++)
    SETIVEC(i, BIOSSEG, ROM_BIOS_IRET);		/* reserved */
  for (i = 0x70; i < 0x77; i++)
    SETIVEC(i, BIOSSEG, ROM_BIOS_IRQ_HI);		/* irq 8-14 */
  SETIVEC(i, BIOSSEG, ROM_BIOS_IRQ15);		/* irq 15 */
  for (i = 0x78; i < 0x80; i++)
    SETIVEC(i, 0x0000, 0x0000);			/* reserved */
  for (i = 0x80; i < 0xf1; i++)
    SETIVEC(i, 0x0000, 0x0000);			/* reserved for BASIC */
  for (i = 0xf1; i < 0x100; i++)
    SETIVEC(i, 0x0000, 0x0000);			/* reserved for user progs */

  /* Most of these will never get used, btw */

  g_printf("BIOS: setting chained BIOS vectors\n");

  int_register_vector(0x08, int08, INT_CHAIN);
  int_register_vector(0x09, int09, INT_CHAIN);
  int_register_vector(0x10, int10, INT_CHAIN);
  int_register_vector(0x0a, int_irq_lo, INT_CHAIN);
  int_register_vector(0x0b, int_irq_lo, INT_CHAIN);
  int_register_vector(0x0c, int_irq_lo, INT_CHAIN);
  int_register_vector(0x0d, int_irq_lo, INT_CHAIN);
  int_register_vector(0x0e, int_irq_lo, INT_CHAIN);
  int_register_vector(0x0f, int_irq_lo, INT_CHAIN);
  int_register_vector(0x10, int10, INT_CHAIN);
  int_register_vector(0x11, int11, INT_CHAIN);
  int_register_vector(0x12, int12, INT_CHAIN);
  int_register_vector(0x13, int13, INT_CHAIN);
  int_register_vector(0x14, int14, INT_CHAIN);
  int_register_vector(0x15, int15, INT_CHAIN);
  int_register_vector(0x16, int16, INT_CHAIN);
  int_register_vector(0x17, int17, INT_CHAIN);
  int_register_vector(0x18, int18, INT_CHAIN);
  int_register_vector(0x19, int19, INT_CHAIN);
  int_register_vector(0x1a, int1a, INT_CHAIN);

  int_register_vector(0x70, int70, INT_CHAIN);
  int_register_vector(0x71, int_irq_hi, INT_CHAIN);
  int_register_vector(0x72, int_irq_hi, INT_CHAIN);
  int_register_vector(0x73, int_irq_hi, INT_CHAIN);
  int_register_vector(0x74, int_irq_hi, INT_CHAIN);
  int_register_vector(0x75, int75, INT_CHAIN);
  int_register_vector(0x76, int_irq_hi, INT_CHAIN);
  int_register_vector(0x77, int77, INT_CHAIN);
}
