/*
 * DANG_BEGIN_MODULE
 *
 * Description:
 *   This module contains everything necessary to set up the BIOS memory
 *   area (the segment 0x0040)
 * 
 * Maintainers: 
 *
 * DANG_END_MODULE
 *
 * $Date: $
 * $Source: $
 * $Revision: $
 * $State: $
 *
 * $Log: $
 */

#include <unistd.h>

#include "dosemu.h"
#include "emu_bios.h"
#include "emu_serial.h"
#include "emu_lpt.h"
#include "emu_mouse.h"

static Bit16u bios_config;

void bios_init_configuration(void)
{
  bios_config = 0;

  /* Configuration word:
   *   bit 0     = IPL diskette installed
   *   bit 1     = math coprocessor
   *   bit 2     = PS/2 pointing device installed
   *   bit 3     = unused
   *   bit 4-5   = initial video mode
   *   bit 6-7   = number of floppy drives installed - 1
   *   bit 8     = 0 if DMA present
   *   bit 9-11  = number of serial ports
   *   bit 12    = game adapter installed
   *   bit 13    = unused, internal modem on PS/2
   *   bit 14-15 = number of printer ports
   */

  if (config.mathco)
    bios_config |= 0x02;

  if ((mice->type == MOUSE_PS2) || mice->intdrv)
    bios_config |= 0x04;

  bios_config |= (config.fdisks - 1) << 6;
  bios_config |= config.num_ser << 9;
  bios_config |= config.num_lpt << 15;
}


void bios_mem_init(void)
{
}

void bios_mem_reset(void)
{
  WRITE_WORD(bios_base_addr_com1, com[0].base_port);
  WRITE_WORD(bios_base_addr_com2, com[1].base_port);
  WRITE_WORD(bios_base_addr_com3, com[2].base_port);
  WRITE_WORD(bios_base_addr_com4, com[3].base_port);

  WRITE_WORD(bios_base_addr_lpt1, lpt[0].base_port);
  WRITE_WORD(bios_base_addr_lpt2, lpt[1].base_port);
  WRITE_WORD(bios_base_addr_lpt3, lpt[2].base_port);

#if 0
  WRITE_WORD(bios_extended_bios, ???);
#endif
  WRITE_WORD(bios_configuration, bios_config);
  WRITE_WORD(bios_memory_size, config.mem_size);

  WRITE_WORD(bios_keyboard_state, 0);    /* no ctrl/shift/alt etc. pressed */
  WRITE_BYTE(bios_keyboard_leds, 0);     /* no keyboard led's set */
#if 0
  WRITE_WORD(bios_keyboard_token, ???);  /* 'storage for alt. keypad entry' */
#endif

  WRITE_WORD(bios_keyboard_buffer_head, 0x001E);
  WRITE_WORD(bios_keyboard_buffer_tail, 0x001E);

#if 0
  WRITE_BYTE(bios_drive_recalib_stat, ???);
  WRITE_BYTE(bios_drive_floppy_stat, ???);
  WRITE_BYTE(bios_motor_shutoff, ???);
  WRITE_BYTE(bios_disk_status, ???);

  /* bios_fdc_result_buffer */

  WRITE_BYTE(bios_video_mode, ???);
  WRITE_WORD(bios_screen_columns, ???);
  WRITE_WORD(bios_video_memory_used, ???);
  WRITE_WORD(bios_video_memory_addr, ???);

  for (i = 0; i < 8; i++) {
    WRITE_BYTE(bios_cursor_x_position(i), ???);
    WRITE_BYTE(bios_cursor_y_position(i), ???);
  }

  WRITE_WORD(bios_cursor_shape, ???);
  WRITE_BYTE(bios_cursor_last_line, ???);
  WRITE_BYTE(bios_cursor_first_line, ???);
  WRITE_BYTE(bios_current_screen_page, ???);
  WRITE_WORD(bios_video_port, ???);
  WRITE_BYTE(bios_vdu_control, ???);
  WRITE_WORD(bios_vdu_control_register, ???);

  WRITE_DWORD(bios_timer, ???);
  WRITE_BYTE(bios_24_hours_flag, 0);

  WRITE_BYTE(bios_keyboard_flags, ???);
#endif

  WRITE_WORD(bios_ctrl_alt_del_flag, 0);

#if 0
  WRITE_BYTE(bios_hdrive_stat, ???);
  WRITE_BYTE(bios_hdrive_count, ???);
  WRITE_BYTE(bios_xt_fdisk_ctrl, ???);
  WRITE_BYTE(bios_hdrive_port_off, ???);
  WRITE_BYTE(bios_lpt1_timeout, ???);
  WRITE_BYTE(bios_lpt2_timeout, ???);
  WRITE_BYTE(bios_lpt3_timeout, ???);
  WRITE_BYTE(bios_lpt4_timeout, ???);
  WRITE_BYTE(bios_com1_timeout, ???);
  WRITE_BYTE(bios_com2_timeout, ???);
  WRITE_BYTE(bios_com3_timeout, ???);
  WRITE_BYTE(bios_com4_timeout, ???);
#endif

  WRITE_WORD(bios_keyboard_buffer_start, 0x001E);
  WRITE_WORD(bios_keyboard_buffer_end,   0x003E);

#if 0
  WRITE_BYTE(bios_rows_on_screen_minus_1, ???);
  WRITE_WORD(bios_font_height, ???);

  WRITE_BYTE(bios_video_info_0, ???);
  WRITE_BYTE(bios_video_info_1, ???);
  WRITE_BYTE(bios_video_info_2, ???);
  WRITE_BYTE(bios_video_combo, ???);

  WRITE_BYTE(bios_floppy_readrate, ???);
  WRITE_BYTE(bios_hdisk_status, ???);
  WRITE_BYTE(bios_hdisk_error, ???);
  WRITE_BYTE(bios_hdisk_busy, ???);
  WRITE_BYTE(bios_combo_disk_card, ???);
  WRITE_BYTE(bios_media_state, ???);
  WRITE_BYTE(bios_floppy0_track, ???);
  WRITE_BYTE(bios_floppy1_track, ???);
#endif

  WRITE_WORD(bios_keyboard_flags2, 16);        /* 102 key keyboard */
  WRITE_DWORD(bios_usr_wait_flg_ptr, 0x00000000);
  WRITE_DWORD(bios_usr_wait_rtc, 0x00000000);
  WRITE_BYTE(bios_rtc_flag, 0);
  WRITE_BYTE(bios_lana_flags, 0);
  WRITE_BYTE(bios_lana_status0, 0);
  WRITE_BYTE(bios_lana_status1, 0);
  WRITE_DWORD(bios_hdisk_int_vec, 0x00000000);
  WRITE_DWORD(bios_video_saveptr, 0xf000f0a4);
  WRITE_WORD(bios_print_screen_flag, 0x00);
}
