/*
 * DANG_BEGIN_MODULE
 * 
 * PHASE0 initialization:
 *
 * This is where you initialize anything related to the UNIX environment.
 * Do version & os detection, at console/not at console, read config info,
 * etc. here
 *
 * DANG_END_MODULE
 * 
 * DANG_BEGIN_CHANGELOG
 *
 * $Date: $
 * $Source: $
 * $Revision: $
 * $State: $
 *
 * $Log: $
 *
 * DANG_END_CHANGELOG
 */

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>

#include "dosemu.h"

extern void pkt_check_receive_quick(void);
extern void SIG_init(void);
extern void emumodule_init(void);

/*
 * DANG_BEGIN_FUNCTION stdio_init
 *
 * description:
 *  Initialize stdio, open debugging output file if user specified one
 * 
 * DANG_END_FUNCTION
 */
void stdio_init(void)
{
  struct stat statout, staterr;

  if(dbg_fd) {
    warn("DBG_FD already set\n");
    return;
  }

  if (config.debugout) {
    dbg_fd = fopen(config.debugout,"w");
    if (!dbg_fd) {
      error("can't open \"%s\" for writing debug file\n", config.debugout);
      exit(1);
    }
  }
  else {
    dbg_fd = 0;
    warn("No debug output file specified, debugging information will not be printed");
  }
  sync();  /* for safety */
  setbuf(stdout, NULL);
}

/*
 * DANG_BEGIN_FUNCTION tmpdir_init
 *
 * description:
 *  Initialize the temporary directory
 * 
 * DANG_END_FUNCTION
 */
void tmpdir_init(void)
{
  /* create tmpdir */
#ifndef RUN_AS_USER
  priv_off();
#endif
  mkdir(tmpdir, S_IREAD | S_IWRITE | S_IEXEC);
#ifndef RUN_AS_USER
  priv_on();
#endif
}

/*
 * DANG_BEGIN_FUNCTION phase0_init
 * 
 * description:
 *  Fire up phase 0
 *
 * DANG_END_FUNCTION
 */
void phase0_init(int argc, char **argv)
{
  if (0 == geteuid()) {
    warn("I am root\n");
    i_am_root = 1;
  } else {
    if (getuid() != geteuid()) {
      fprintf(stderr, "Can't setuid to anyone else but root\n");
      exit(1);
    }
  }
  
#ifdef RUN_AS_USER
  /* start running as real, not effective user.  This is moved from */
  /* stdio_init, as that and config_init have now changed places and */
  /* allowing reading arbitrary files while being root is not smart */
  
  priv_off();	
#endif

  /* the transposal of (config_|stdio_)init allows the addition of -o */
  /* to specify a debug out filename, if you're wondering */

  config_init(argc, argv);	/* parse the commands & config file(s) */
  stdio_init();		        /* initialize stdio & open debug file  */
  version_init();               /* what linux version are we running?  */
  emumodule_init();
  SIG_init();
  tmpdir_init();

  /* initialize the internal tables */
  int_init();
  port_init();
  memcheck_init();
  timer_init();
  signal_init();

  env_init_phase0();
}



