/* 
 * DANG_BEGIN_MODULE
 * 
 * Contains the run_vm86 function
 *
 * DANG_END_MODULE
 *
 * DANG_BEGIN_CHANGELOG
 *
 * $Date: $
 * $Source: $
 * $Revision: $
 * $State: $
 *
 * $Log: $
 *
 * DANG_END_CHANGELOG
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>

#include "dosemu.h"
#include "emu_mem.h"
#include "emu_cpu.h"
#include "emu_port.h"
#include "emu_int.h"
#include "emu_pic.h"
#include "emu_port.h"
#include "emu_hlt.h"
#include "emu_timers.h"

/*
 * DANG_BEGIN_FUNCTION vm86_GP_fault();
 *
 * description:
 * All from the kernel unhandled general protection faults from V86 mode
 * are handled here. This are mainly port IO and the HLT instruction.
 *
 * DANG_END_FUNCTION
 */

void vm86_GP_fault(void)
{
  Bit8u *csp;

  if (ignore_segv) {
    error("ERROR: sigsegv ignored!\n");
    return;
  }

  if (in_sigsegv)
    error("ERROR: in_sigsegv=%d!\n", in_sigsegv);

  /* in_vm86 = 0; */
  in_sigsegv++;

  /* DANG_BEGIN_REMARK 
   * In a properly functioning emulator :-), sigsegv's will never come
   * while in a non-reentrant system call (ioctl, select, etc).  Therefore,
   * there's really no reason to worry about them, so I say that I'm NOT
   * in a signal handler (I might make this a little clearer later, to
   * show that the purpose of in_sighandler is to stop non-reentrant system
   * calls from being reentered.
   * I reiterate: sigsegv's should only happen when I'm running the vm86
   * system call, so I really shouldn't be in a non-reentrant system call
   * (except maybe vm86)
   * - Robert Sanders
   * DANG_END_REMARK
   */
  in_sighandler = 0;

  if (LWORD(eflags) & TF) {
    g_printf("SIGSEGV received while TF is set\n");
    show_regs(__FILE__, __LINE__);
  }

  csp = SEG_ADR((Bit8u *), cs, ip);

  switch (*csp) {

  case 0x6c:                    /* insb */
    *(SEG_ADR((unsigned char *),es,di)) = port_inb((int) LWORD(edx));
    i_printf("insb(0x%04x) value %02x\n",
            LWORD(edx), *(SEG_ADR((unsigned char *),es,di)));   
    if (LWORD(eflags) & DF) LWORD(edi)--;
    else LWORD(edi)++;
    LWORD(eip)++;
    break;

  case 0x6d:			/* insw */
    *(SEG_ADR((unsigned short *),es,di)) = port_inw((int) LWORD(edx));
    i_printf("insw(0x%04x) value %04x \n",
            LWORD(edx),*(SEG_ADR((unsigned short *),es,di)));
    if (LWORD(eflags) & DF) LWORD(edi) -= 2;
    else LWORD(edi) +=2;
    LWORD(eip)++;
    break;
      

  case 0x6e:			/* outsb */
    fprintf(stderr,"untested: outsb port 0x%04x value %02x\n",
            LWORD(edx),*(SEG_ADR((unsigned char *),ds,si)));
    port_outb(LWORD(edx), *(SEG_ADR((unsigned char *),ds,si)));
    if (LWORD(eflags) & DF) LWORD(esi)--;
    else LWORD(esi)++;
    LWORD(eip)++;
    break;


  case 0x6f:			/* outsw */
    fprintf(stderr,"untested: outsw port 0x%04x value %04x\n",
            LWORD(edx), *(SEG_ADR((unsigned short *),ds,si)));
    port_outw(LWORD(edx), *(SEG_ADR((unsigned short *),ds,si)));
    if (LWORD(eflags) & DF ) LWORD(esi) -= 2;
    else LWORD(esi) +=2; 
    LWORD(eip)++;
    break;

  case 0xe5:			/* inw xx */
    LWORD(eax) = port_inw((int) csp[1]);
    LWORD(eip) += 2;
    break;
  case 0xe4:			/* inb xx */
    LWORD(eax) &= ~0xff;
    LWORD(eax) |= port_inb((int) csp[1]);
    LWORD(eip) += 2;
    break;

  case 0xed:			/* inw dx */
    LWORD(eax) = port_inw(LWORD(edx));
    LWORD(eip) += 1;
    break;
  case 0xec:			/* inb dx */
    LWORD(eax) &= ~0xff;
    LWORD(eax) |= port_inb(LWORD(edx));
    LWORD(eip) += 1;
    break;

  case 0xe7:			/* outw xx */
    port_outb((int) csp[1], LO(ax));
    port_outb((int) csp[1] + 1, HI(ax));
    LWORD(eip) += 2;
    break;
  case 0xe6:			/* outb xx */
    port_outb((int) csp[1], LO(ax));
    LWORD(eip) += 2;
    break;

  case 0xef:			/* outw dx */
    port_outb(REG(edx), LO(ax));
    port_outb(REG(edx) + 1, HI(ax));
    LWORD(eip) += 1;
    break;
  case 0xee:			/* outb dx */
    port_outb(LWORD(edx), LO(ax));
    LWORD(eip) += 1;
    break;

  case 0xf3:                    /* rep */
    switch(csp[1] & 0xff) {                
      case 0x6c: {             /* rep insb */
        int delta = 1;
        if(LWORD(eflags) &DF ) delta = -1;
        i_printf("Doing REP F3 6C (rep insb) %04x bytes, DELTA %d\n",
                LWORD(ecx),delta);
        while (LWORD(ecx))  {
          *(SEG_ADR((unsigned char *),es,di)) = port_inb((int) LWORD(edx));
          LWORD(edi) += delta;
          LWORD(ecx)--;
        }
        LWORD(eip)+=2;
        break;
      }
      case  0x6d: {           /* rep insw */
        int delta =2;
        if(LWORD(eflags) & DF) delta = -2;
        i_printf("REP F3 6D (rep insw) %04x words, DELTA %d\n",
                LWORD(ecx),delta);
        while(LWORD(ecx)) {
          *(SEG_ADR((unsigned short *),es,di))=port_inw(LWORD(edx));
          LWORD(edi) += delta;
          LWORD(ecx)--;
         }
         LWORD(eip) +=2;
         break;
      }
      case 0x6e: {           /* rep outsb */
        int delta = 1;
        fprintf(stderr,"untested: rep outsb\n");
        if(LWORD(eflags) & DF) delta = -1;
        while(LWORD(ecx)) {
          port_outb(LWORD(edx), *(SEG_ADR((unsigned char *),ds,si)));
          LWORD(esi) += delta;
          LWORD(ecx)--;
        }
        LWORD(eip)+=2;
        break;
      }
      case 0x6f: { 
        int delta = 2;
        fprintf(stderr,"untested: rep outsw\n");
        if(LWORD(eflags) & DF) delta = -2;
        while(LWORD(ecx)) {
          port_outw(LWORD(edx), *(SEG_ADR((unsigned short *),ds,si)));
          LWORD(esi) += delta;
          LWORD(ecx)--;
        }
        LWORD(eip)+=2;
        break;
      }
      default:
      fprintf(stderr, "Nope REP F3,CSP[1] = 0x%04x\n", csp[1]);
    }
    break;        

  case 0xf4:
    hlt_handle();
    break;

  case 0xf0:			/* lock */
  default:
    /* er, why don't we advance eip here, and
	 why does it work??  */
    error("general protection at %p: %x\n", csp,*csp);
    show_regs(__FILE__, __LINE__);
    show_ints(0, 0x33);
    error("ERROR: SIGSEGV, protected insn...exiting!\n");
    fatalerr = 4;
    leavedos(fatalerr);		/* shouldn't return */
  }				/* end of switch() */

  if (LWORD(eflags) & TF) {
    g_printf("TF: trap done");
    show_regs(__FILE__, __LINE__);
  }

  in_sigsegv--;
  in_sighandler = 0;
}

/*
 * DANG_BEGIN_FUNCTION run_vm86
 * 
 * description: Here is where DOSEMU runs VM86 mode with the vm86() call
 * which also has the registers that it will be called with.  It will stop
 * vm86 mode for many reasons, like trying to execute an interrupt, doing
 * port I/O to ports not opened for I/O, etc ...
 * 
 * DANG_END_FUNCTION
 */
void cpu_loop(void)
{
  struct timeval cur_time;
  int retval;

  in_vm86 = TRUE;
  retval  = vm86(&vm86s);
  in_vm86 = FALSE;

  switch (VM86_TYPE(retval)) {
    case VM86_UNKNOWN:
      vm86_GP_fault();
      break;
    case VM86_STI:
      break;
    case VM86_INTx:
      int_run(VM86_ARG(retval));
      break;
    case VM86_SIGNAL:
      I_printf("Return for SIGNAL\n");
      break;
    default:
      error("unknown return value from vm86()=%x,%d-%x\n", VM86_TYPE(retval),
	    VM86_TYPE(retval), VM86_ARG(retval));
      fatalerr = 4;
    }

  gettimeofday(&cur_time, NULL);

  handle_signals();
  timer_schedule(&cur_time);

  /*
   * This is here because ioctl() is non-reentrant, and signal handlers
   * may have to use ioctl().  This results in a possible (probable)
   * time lag of indeterminate length (and a bad return value). Ah, life
   * isn't perfect.
   * 
   * I really need to clean up the queue functions to use real queues.
   */
  if (iq.queued)
    do_queued_ioctl();

  if (__interrupts_enabled__) {
    if (pic_queued()) {
      pic_do_irq();
    }
  }
}
