/*
 * $Date: $
 * $Source: $
 * $Revision: $
 * $State: $
 *
 * $Log: $
 */

#ifndef _DOSEMU_H
#define _DOSEMU_H

#include <unistd.h>
#include <stdio.h>
#include <sys/types.h>

#include "emu_defs.h"
#include "emu_cpu.h"
#include "emu_debug.h"
#include "emu_signal.h"

#ifdef CONFIG_EMUMODULE
  #include "arch/emusys.h"
#endif

#if CONFIG_SIG
typedef struct { int fd; int irq; } SillyG_t;
#endif

/*
 * DANG_BEGIN_REMARK
 * The `vm86_struct` is used to pass all the necessary status/registers to
 * DOSEMU when running in vm86 mode.
 * DANG_END_REMARK
 */ 
EXTERN struct vm86_struct vm86s;
EXTERN int terminal_pipe;
EXTERN int terminal_fd INIT(-1);

/* set to one if runing setuid as root */
EXTERN Boolean  i_am_root INIT(FALSE);

EXTERN char *tmpdir;

/* number of highest vid page - 1 */
EXTERN int max_page INIT(7);

/* the fd for the keyboard */ 
EXTERN int kbd_fd INIT(-1);

/* the file descriptor for /dev/mem when mmap'ing the video mem */
EXTERN int mem_fd INIT(-1);

/* X-pipes */
EXTERN int keypipe;
EXTERN int mousepipe;

EXTERN Boolean in_vm86 INIT(FALSE);

#define NOWAIT  0
#define WAIT    1
#define TEST    2
#define POLL    3

struct ioctlq {
  int fd, req, param3;
  int queued;
};

/* one-entry queue ;-( for ioctl's */
EXTERN struct ioctlq iq INIT({0, 0, 0, 0}); 
EXTERN u_char in_ioctl INIT(0);
EXTERN struct ioctlq curi INIT({0, 0, 0, 0});

/* would like to have this in memory.h (but emu.h is included before memory.h !) */
#define HARDWARE_RAM_START 0xc8000
#define HARDWARE_RAM_STOP  0xf0000

typedef struct config_info {
  int    hdiskboot;

  /* for video */
  Boolean  console;
  Boolean  console_video;
  Boolean  graphics;
  Boolean  vga;
  Boolean  X;
  Bit16u   cardtype;
  Bit16u   chipset;
  Bit16u   gfxmemsize;		/* for SVGA card, in K */

  Bit16u   term_color;		/* Terminal color support on or off */
  Bit16u   term_updatefreq;	/* Terminal update frequency */
  Bit16u   term_charset;		/* Terminal Character set */
  Bit16u   term_esc_char;	        /* ASCII value used to get slang help screen */
  Bit16u   X_updatelines;         /* Amount to update at a time */
  Bit16u   X_updatefreq;          /* X update frequency */
  char    *X_display;           /* X server to use (":0") */
  char    *X_title;               /* X window title */
  char    *X_icon_name;
  char    *X_font;
  int      X_blinkrate;
  Boolean  fullrestore;
  Boolean  dualmon;
  
  Bit16u   usesX;  /* !=0 if dosemu owns an X window */
  
  Boolean  console_keyb;
  Boolean  kbd_tty;
  Boolean  X_keycode;	/* use keycode field of event structure */
  Boolean  exitearly;
  Boolean  mathco;
  Boolean  ipxsup;
  Boolean  keybint;
  Boolean  dosbanner;
  Boolean  allowvideoportaccess;
  Boolean  timers;
  Boolean  mouse_flag;
  Boolean  mapped_bios;	/* video BIOS */
  Boolean  mapped_sbios;	/* system BIOS */
  char    *vbios_file;	/* loaded VBIOS file */
  Boolean  vbios_copy;
  Bit16u   vbios_seg;           /* VGA-BIOS-segment for mapping */
  Bit16u   vbios_size;          /* size of VGA-BIOS (64K for vbios_seg=0xe000
				   32K for vbios_seg=0xc000) */
  
  Boolean  bootdisk;	/* Special bootdisk defined */
  int      fastfloppy;
  char    *emusys;		/* map CONFIG.SYS to CONFIG.EMU */
  char    *emubat;		/* map AUTOEXEC.BAT to AUTOEXEC.EMU */
  
  Bit16u   speaker;		/* 0 off, 1 native, 2 emulated */
  Bit16u   fdisks, hdisks;
  Bit16u   num_lpt;
  Bit16u   num_ser;
  Bit16u   num_mice;
  
  int      pktflags;		/* global flags for packet driver */
  
  Bit16u   hogthreshold;
  
  Bit16u   mem_size, xms_size, ems_size, dpmi, max_umb;
  Bit16u   ems_frame;
  Boolean  must_spare_hardware_ram;
  Boolean  hardware_pages[ ((HARDWARE_RAM_STOP-HARDWARE_RAM_START) >> 12)+1 ];
    
  int sillyint;            /* IRQ numbers for Silly Interrupt Generator 
			      (bitmask, bit3..15 ==> IRQ3 .. IRQ15) */
  
  int keyboard;
  unsigned char *key_map;     /* pointer to the correct keyboard-map */
  unsigned char *shift_map;
  unsigned char *alt_map;
  unsigned char *num_table;
  unsigned short detach;
  unsigned char *debugout;
  
  /* Lock File business */
  char    *tty_lockdir;  	/* The Lock directory  */
  char    *tty_lockfile;	/* Lock file pretext ie LCK.. */
  Boolean  tty_lockbinary;	/* Binary lock files ? */
  
} config_t;

#define SPKR_OFF	0
#define SPKR_NATIVE	1
#define SPKR_EMULATED	2

/* 
 * DANG_BEGIN_REMARK
 * DOSEMU keeps system wide configuration status in a structure
 * called config.
 * DANG_END_REMARK
 */
EXTERN struct config_info config;

/*
 * DANG_BEGIN_REMARK
 * The var `fatalerr` can be given a true value at any time to have DOSEMU
 * exit on the next return from vm86 mode.
 * DANG_END_REMARK
 */
EXTERN int fatalerr;

extern int   mfs_redirector(void);
extern int   mfs_inte6(void);
extern void  io_select(fd_set);
extern void  open_kmem(void);
extern void  close_kmem(void);
extern int   parse_config(char *);
extern void  pkt_helper(void);
extern short pop_word(struct vm86_regs *);
extern void  ems_init(void);
extern void  getKeys(void);
extern void  keybuf_clear(void);
extern void  saytime(char *m_str);
extern void  p_dos_str(char *,...) FORMAT(printf, 1, 2);
extern void  char_out(u_char, int);
extern int   set_ioperm(int, int, int);

extern void  leavedos(int) NORETURN;

extern void  dos_ctrlc(void);
extern void  dos_ctrl_alt_del(void);

extern void  do_queued_ioctl(void);
extern int   queue_ioctl(int, int, int);
extern int   do_ioctl(int, int, int);

extern void  add_to_io_select(int, unsigned char);
extern void  remove_from_io_select(int, unsigned char);

#endif /* EMU_H */
