/*
 * general constants/functions for memory, addresses, etc.
 *    for Linux DOS emulator, Robert Sanders, gt8134b@prism.gatech.edu
 *
 * $Date: $
 * $Source: $
 * $Revision: $
 * $State: $
 *
 * $Log: $
 */

#ifndef _EMU_MEM_H
#define _EMU_MEM_H

#ifdef __cplusplus
#define asmlinkage extern "C"
#else
#define asmlinkage
#endif

#ifdef __ELF__
#ifdef __STDC__
#define CISH_INLINE(x) #x
#define CISH(x) x
#else
#define CISH_INLINE(x) "x"
#define CISH(x) x
#endif
#else
#ifdef __STDC__
#define CISH_INLINE(x) "_"#x
#define CISH(x) _##x
#else
#define CISH_INLINE(x) "_x"
#define CISH(x) _/**/x
#endif
#endif

#ifndef BIOSSEG
#define BIOSSEG		0xf000
#endif

#ifdef CONFIG_OLD_BIOS

/* split segment 0xf000 into two region, 0xf0000 to 0xf7fff is read-write */
/*                                       0xf8000 to 0xfffff is read-only  */
/* so put anything need read-write into BIOSSEG and anything read-only */
/* to ROMBIOSSEG  */

#define INT09_SEG	BIOSSEG
#define INT09_OFF	0xe987		/* for 100% IBM compatibility */
#define INT09_ADD	((INT09_SEG << 4) + INT09_OFF)

/* The packet driver has some code in this segment which needs to be */
/* at BIOSSEG.  therefore use BIOSSEG and compensate for the offset. */
/* Memory required is about 2000 bytes, beware! */
#define PKTDRV_SEG	(BIOSSEG)
#define PKTDRV_OFF	0x3100
#define PKTDRV_ADD	((PKTDRV_SEG << 4) + PKTDRV_OFF)

/* don't change these for now, they're hardwired! */
#define Mouse_SEG       (BIOSSEG)
#define Mouse_OFF       0x20f0
#define Mouse_ROUTINE_OFF  0x2140
#define Mouse_ADD      ((Mouse_SEG << 4)+Mouse_OFF)
#define Mouse_ROUTINE  ((Mouse_SEG << 4)+Mouse_ROUTINE_OFF)

#ifndef ROMBIOSSEG
#define ROMBIOSSEG	0xf800
#endif

#define XMSControl_SEG  ROMBIOSSEG
#define XMSControl_OFF  0
#define XMSControl_ADD  ((XMSControl_SEG << 4)+XMSControl_OFF)
#define XMSTrap_ADD     ((XMSControl_SEG << 4)+XMSControl_OFF+5)

/* EMS origin must be at 0 */
#define EMS_SEG		(ROMBIOSSEG+0x100)
#define EMS_OFF		0x0000
#define EMS_ADD		((EMS_SEG << 4) + EMS_OFF)

#define Banner_SEG	ROMBIOSSEG
#define Banner_OFF	0x2000
#define Banner_ADD	((Banner_SEG << 4) + Banner_OFF)

#define INT16_SEG	ROMBIOSSEG
#define INT16_OFF	0x3000
#define INT16_ADD	((INT16_SEG << 4) + INT16_OFF)

#define IPX_SEG		ROMBIOSSEG
#define IPX_OFF		0x3100
#define IPX_ADD		((IPX_SEG << 4) + IPX_OFF)

#define INT08_SEG	ROMBIOSSEG
#define INT08_OFF	0x4000
#define INT08_ADD	((INT08_SEG << 4) + INT08_OFF)

#define INT10_SEG	ROMBIOSSEG
#define INT10_OFF	0x4200
#define INT10_ADD	((INT10_SEG << 4) + INT10_OFF)

/* This inline interrupt is used for FCB open calls */
#define INTE7_SEG	ROMBIOSSEG
#define INTE7_OFF	0x4500
#define INTE7_ADD	((INTE7_SEG << 4) + INTE7_OFF)

#define PIC_SEG       ROMBIOSSEG
#define PIC_OFF       0x47f0
#define PIC_ADD       ((PIC_SEG << 4) + PIC_OFF)

#define DPMI_SEG	ROMBIOSSEG
#define DPMI_OFF	0x4800		/* need at least 512 bytes */
#define DPMI_ADD	((DPMI_SEG << 4) + DPMI_OFF)

/* For int15 0xc0 */
#define ROM_CONFIG_SEG  BIOSSEG
#define ROM_CONFIG_OFF  0xe6f5
#define ROM_CONFIG_ADD	((ROM_CONFIG_SEG << 4) + ROM_CONFIG_OFF)

#endif

#define EMM_BASE_ADDRESS        (config.ems_frame << 4)
#define EMM_SEGMENT             (config.ems_frame)

#define LASTSCAN_SEG	(BIOSSEG + 0x100)
#define LASTSCAN_OFF	0x0
#define LASTSCAN_ADD	(u_short *)((LASTSCAN_SEG << 4) + LASTSCAN_OFF)

#define OUTB_SEG	(BIOSSEG+0x100)
#define OUTB_OFF	0x1000
#define OUTB_ADD	(u_short *)((OUTB_SEG << 4) + OUTB_OFF)

#define VBIOS_START	(config.vbios_seg << 4 )
/*#define VBIOS_SIZE	(64*1024)*/
#define VBIOS_SIZE	(config.vbios_size)
#define GFX_CHARS	0xffa6e
#define GFXCHAR_SIZE	1400

/* Memory adresses for all common video adapters */

#define MDA_PHYS_TEXT_BASE  0xB0000
#define MDA_VIRT_TEXT_BASE  0xB0000

#define CGA_PHYS_TEXT_BASE  0xB8000
#define CGA_VIRT_TEXT_BASE  0xB8000

#define EGA_PHYS_TEXT_BASE  0xB8000
#define EGA_VIRT_TEXT_BASE  0xB8000

#define VGA_PHYS_TEXT_BASE  0xB8000
#define VGA_VIRT_TEXT_BASE  0xB8000

#define CO      80 /* A-typical screen width */
#define LI      25 /* Normal rows on a screen */
#define TEXT_SIZE	(li*co*2) /* 4096 text page size */
#define PAGE_ADDR(pg)	(caddr_t)(virt_text_base + (pg*TEXT_SIZE))
#define SCREEN_ADR(s)	((us *)(virt_text_base + (s*TEXT_SIZE)))

#define GRAPH_BASE 0xA0000
#define GRAPH_SIZE 0x20000

#define BIOS_DATA_SEG   (0x400)	/* for absolute adressing */
#define BIOS_DATA_PTR(off) ((void *) (BIOS_DATA_SEG + off))

#define KBD_START BIOS_DATA_PTR(0x80)
#define KBD_END   BIOS_DATA_PTR(0x82)
#define KBD_HEAD  BIOS_DATA_PTR(0x1A)
#define KBD_TAIL  BIOS_DATA_PTR(0x1C)

#define KBDFLAG_ADDR (unsigned short *)BIOS_DATA_PTR(0x17)
#define KEYFLAG_ADDR (unsigned short *)BIOS_DATA_PTR(0x96)


#ifndef CONFIG_SCOTT
/*
 * The following are the non-standard "dosemu bios" addresses
 */
#define ROM_Mouse_OFF       0xe2d0   /* internal mouse driver */
#define ROM_Mouse_ROUTINE_OFF 0xe2e0 /* ??? internal mouse driver */
#define ROM_XMSControl_SEG  0xfe2f   /* Does XMS have to be at f800:0000? */
#define ROM_XMSControl_OFS  0x0000
#define ROM_PIC_HLT         0xe2ff   /* Offset of 1 byte hlt */
#define ROM_Banner          0xe300   /* Banner code */
#define ROM_IPX             0xe310   /* IPX code */
#define ROM_Video_Bootup    0xe330   /* video bootup code */
#define ROM_FCB_Handler     0xe350   /* int e7 FCB handler */
#define ROM_DPMI_Offset     0xf100   /* currently located on it's own */
                                     /* reserved through 0xf500 */
#else
#define ROM_Mouse_OFF       0xa2d0   /* internal mouse driver */
#define ROM_Mouse_ROUTINE_OFF 0xa2e0 /* ??? internal mouse driver */
#define ROM_XMSControl_SEG  0xfa2f   /* Does XMS have to be at f800:0000? */
#define ROM_XMSControl_OFS  0x0000
#define ROM_PIC_HLT         0xa2ff   /* Offset of 1 byte hlt */
#define ROM_Banner          0xa300   /* Banner code */
#define ROM_IPX             0xa310   /* IPX code */
#define ROM_Video_Bootup    0xa330   /* video bootup code */
#define ROM_FCB_Handler     0xa350   /* int e7 FCB handler */
#define ROM_DPMI_Offset     0xb100   /* currently located on it's own */
                                     /* reserved through 0xf500 */
#endif


#define Mouse_SEG          ((Bit32u)BIOSSEG)
#define Mouse_OFF          ((Bit32u)ROM_Mouse_OFF)
#define Mouse_ROUTINE_OFF  ((Bit32u)ROM_Mouse_ROUTINE_OFF)
#define Mouse_ADD          ((Mouse_SEG << 4)+Mouse_OFF)
#define Mouse_ROUTINE      ((Mouse_SEG << 4)+Mouse_ROUTINE_OFF)
#define PIC_SEG            ((Bit32u)BIOSSEG)
#define PIC_OFF            ((Bit32u)ROM_PIC_HLT)
#define PIC_ADD            ((PIC_SEG << 4) + PIC_OFF)
#define XMSControl_SEG     ((Bit32u)ROM_XMSControl_SEG)
#define XMSControl_OFF     ((Bit32u)ROM_XMSControl_OFS)
#define XMSControl_ADD     ((XMSControl_SEG << 4)+XMSControl_OFF)
#define XMSTrap_ADD        ((XMSControl_SEG << 4)+XMSControl_OFF+5)
#define ROM_CONFIG_SEG     BIOSSEG
#define ROM_CONFIG_OFF     ROM_BIOS_SYSPARM
#define ROM_CONFIG_ADD	   ((ROM_CONFIG_SEG << 4) + ROM_CONFIG_OFF)
#define DPMI_SEG	   BIOSSEG
#define DPMI_OFF	   ROM_DPMI_Offset
#define DPMI_ADD           ((DPMI_SEG << 4) + DPMI_OFF)
#define INT09_SEG	   BIOSSEG
#define INT09_OFF	   ROM_BIOS_INT09
#define INT09_ADD          ((INT09_SEG << 4) + INT09_OFF)
#define INTE7_SEG	   BIOSSEG
#define INTE7_OFF	   ROM_FCB_Handler
#define INTE7_ADD	   ((INTE7_SEG << 4) + INTE7_OFF)




#ifndef CONFIG_SCOTT
/*
 * The following are (to the best of my knowledge) the standard BIOS
 * entry points and table addresses
 */
#define ROM_BIOS_IRQ_LO     0xe2a0   /* irq 2-7 handler */
#define ROM_BIOS_IRQ_HI     0xe2b0   /* irq 8-15 handler */
#define ROM_BIOS_SELFTEST   0xe05b   /* power on self test addr */
#define ROM_BIOS_BOOTUP     0xe070   /* our BOOT code */
#define ROM_BIOS_NMI        0xe2c3   /* nmi handler */
#define ROM_BIOS_INT13HARD  0xe3fe   /* int 13 hard drive handler */
#define ROM_BIOS_HARDPARM   0xe401   /* hard drive param. table */
#define ROM_BIOS_HARDPARM2  0xe6b1   /* ??? hard drive param. table */
#define ROM_BIOS_INT19      0xe6f2   /* boot-load entry point */
#define ROM_BIOS_SYSPARM    0xe6f5   /* configuration data table */
#define ROM_BIOS_BAUDRATE   0xe729   /* baud rate generator table */
#define ROM_BIOS_INT14      0xe739   /* int 14 serial handler */
#define ROM_BIOS_INT16      0xe82e   /* int 16 keyboard handler */
#define ROM_BIOS_INT09      0xe987   /* int 09 keyboard interrupt */
#define ROM_BIOS_INT13FLOP  0xec59   /* int 13 floppy handler */
#define ROM_BIOS_INT13      0xec59   /* int 13 entry has to be somewhere... */
#define ROM_BIOS_INT0E      0xef57   /* int 0e floppy driver (irq6) */
#define ROM_BIOS_FLOPPARM   0xefc7   /* diskette controller parms */
#define ROM_BIOS_INT17      0xefd2   /* int 17 printer handler */
#define ROM_BIOS_INT10LO    0xf045   /* int 10 0x00-0x0f handler */
#define ROM_BIOS_INT10      0xf065   /* int 10 handler */
#define ROM_BIOS_VIDEOPARM  0xf0a4   /* CGA/MDA video param table (int 1d) */
#define ROM_BIOS_INT12      0xf841   /* int 12 mem. size handler */
#define ROM_BIOS_INT11      0xf84d   /* int 11 equip. list handler */
#define ROM_BIOS_INT15      0xf859   /* int 15 system services */
#define ROM_BIOS_8x8FONT    0xfa6e   /* 8x8 font */
#define ROM_BIOS_INT1A      0xfe6e   /* int 0x1a time of day handler */
#define ROM_BIOS_INT08      0xfea5   /* int 0x08 timer tick handler */
#define ROM_BIOS_INITINTS   0xfef3   /* initial interrupt vector table */
#define ROM_BIOS_IRQ15      0xff43   /* irq15 code */
#define ROM_BIOS_IRET       0xff53   /* dummy iret instr. */
#define ROM_BIOS_INT05      0xff54   /* int 05 print screen handler */
#define ROM_BIOS_POWERON    0xfff0   /* power-on entry point */
#define ROM_BIOS_DATE       0xfff5   /* ascii rom date mm/dd/yy */
#define ROM_BIOS_MODELID    0xfffe   /* system model id */
#else
#define ROM_BIOS_IRQ_LO     0xa2a0   /* irq 2-7 handler */
#define ROM_BIOS_IRQ_HI     0xa2b0   /* irq 8-15 handler */
#define ROM_BIOS_SELFTEST   0xa05b   /* power on self test addr */
#define ROM_BIOS_BOOTUP     0xa070   /* our BOOT code */
#define ROM_BIOS_NMI        0xa2c3   /* nmi handler */
#define ROM_BIOS_INT13HARD  0xa3fe   /* int 13 hard drive handler */
#define ROM_BIOS_HARDPARM   0xa401   /* hard drive param. table */
#define ROM_BIOS_HARDPARM2  0xa6b1   /* ??? hard drive param. table */
#define ROM_BIOS_INT19      0xa6f2   /* boot-load entry point */
#define ROM_BIOS_SYSPARM    0xa6f5   /* configuration data table */
#define ROM_BIOS_BAUDRATE   0xa729   /* baud rate generator table */
#define ROM_BIOS_INT14      0xa739   /* int 14 serial handler */
#define ROM_BIOS_INT16      0xa82e   /* int 16 keyboard handler */
#define ROM_BIOS_INT09      0xa987   /* int 09 keyboard interrupt */
#define ROM_BIOS_INT13FLOP  0xac59   /* int 13 floppy handler */
#define ROM_BIOS_INT13      0xac59   /* int 13 entry has to be somewhere... */
#define ROM_BIOS_INT0E      0xaf57   /* int 0e floppy driver (irq6) */
#define ROM_BIOS_FLOPPARM   0xafc7   /* diskette controller parms */
#define ROM_BIOS_INT17      0xafd2   /* int 17 printer handler */
#define ROM_BIOS_INT10LO    0xb045   /* int 10 0x00-0x0f handler */
#define ROM_BIOS_INT10      0xb065   /* int 10 handler */
#define ROM_BIOS_VIDEOPARM  0xb0a4   /* CGA/MDA video param table (int 1d) */
#define ROM_BIOS_INT12      0xb841   /* int 12 mem. size handler */
#define ROM_BIOS_INT11      0xb84d   /* int 11 equip. list handler */
#define ROM_BIOS_INT15      0xb859   /* int 15 system services */
#define ROM_BIOS_8x8FONT    0xba6e   /* 8x8 font */
#define ROM_BIOS_INT1A      0xbe6e   /* int 0x1a time of day handler */
#define ROM_BIOS_INT08      0xbea5   /* int 0x08 timer tick handler */
#define ROM_BIOS_INITINTS   0xbef3   /* initial interrupt vector table */
#define ROM_BIOS_IRQ15      0xbf43   /* irq15 code */
#define ROM_BIOS_IRET       0xbf53   /* dummy iret instr. */
#define ROM_BIOS_INT05      0xbf54   /* int 05 print screen handler */
#define ROM_BIOS_POWERON    0xbff0   /* power-on entry point */
#define ROM_BIOS_DATE       0xbff5   /* ascii rom date mm/dd/yy */
#define ROM_BIOS_MODELID    0xbffe   /* system model id */
#endif


#ifndef __ASM__
/* memcheck memory conflict finder definitions */
int  memcheck_addtype(unsigned char map_char, char *name);
void memcheck_reserve(unsigned char map_char, int addr_start, int size);
void memcheck_init(void);
int  memcheck_isfree(int addr_start, int size);
int  memcheck_findhole(int *start_addr, int min_size, int max_size);
void memcheck_dump(void);
#endif

#endif /* _EMU_MEM_H */
